# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name, unused-argument

"""
yabt External command builder
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from ..extend import (
    PropType as PT, register_build_func, register_builder_sig,
    register_manipulate_target_hook)
from ..logging import make_logger
from ..utils import yprint


logger = make_logger(__name__)


register_builder_sig(
    'ExtCommand',
    [('cmd'),
     ('in_buildenv', PT.Target, None),
     ('cmd_env', None),
     ('work_dir', PT.str, None),
     ('auto_uid', PT.bool, True),
     ],
    # not cachable by default, since in general this builder doesn't know how
    # to keep track of its artifacts
    cachable=False)


@register_build_func('ExtCommand')
def ext_command_builder(build_context, target):
    yprint(build_context.conf, 'Build (run) ExtCommand', target)
    build_context.run_in_buildenv(
        target.props.in_buildenv, target.props.cmd, target.props.cmd_env,
        target.props.work_dir, target.props.auto_uid)
    # TODO(itamar): way to describe the artifacts of the external command,
    # so it can be used by dependent targets, and cached in some smart way


@register_manipulate_target_hook('ExtCommand')
def ext_command_manipulate_target(build_context, target):
    target.buildenv = target.props.in_buildenv



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target graph
~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

from collections import defaultdict
from os.path import relpath

import difflib
import networkx
from networkx.algorithms import dag

from .buildfile_parser import process_build_file
from .compat import walk
from .config import Config
from .logging import make_logger
from .target_utils import (
    norm_name, parse_target_selectors, split, split_build_module)

logger = make_logger(__name__)


def build_target_dep_graph(build_context, unused_conf: Config):
    build_context.target_graph = networkx.DiGraph()
    for target_name, target in build_context.targets.items():
        build_context.target_graph.add_node(target_name)
        for dep in target.deps:
            build_context.target_graph.add_edge(target_name, dep)


def norm_rel_target(target_spec, build_module):
    if ':' not in target_spec:
        target_spec += ':*'
    return norm_name(build_module, target_spec)


def generate_all_targets(conf: Config):
    # TODO(itamar): add ignore marker files / flags
    for root, unused_dirs, files in walk(conf.project_root):
        if conf.build_file_name in files:
            yield norm_rel_target(relpath(root, conf.project_root), '//')


class SeedRef:
    """Helper class for tracking references to targets.

    This class is used mainly for error reporting.
    A single instance represents a single target observed during build graph
    processing (AKA "seed"), in any context, whether valid or not (not
    including syntax error, which are caught and reported during YBuild
    parsing - see buildfile_parser module).

    Currently supprted contexts:
    - `on_cli` - whether this seed was specified on the command line (boolean).
    - `from_default` - whether this seed was specified as one of the "default"
      targets that seed the graph if nothing is specified on the command line.
    - `dep_of` - a set of other seeds that referred to this seed as their
      dependency.
    - `buildenv_of` - a set of other seeds that referred to this seed as their
      buildenv.

    Usage: Through a defaultdict of SeedRefs (mapping from seed name to its
    SeedRef instance) in populate_targets_graph - for every target processed,
    need to add the processed target to all the seeds that that target refers
    to (a reverse index).
    When an error is found, the raise_unresolved_targets function may use the
    mapping to report where an invalid seed was observed.
    """

    def __init__(self):
        self.on_cli = False
        self.from_default = False
        self.dep_of = set()
        self.buildenv_of = set()


def raise_unresolved_targets(build_context, conf, unknown_seeds, seed_refs):
    """Raise error about unresolved targets during graph parsing."""

    def format_target(target_name):
        # TODO: suggest similar known target names
        build_module = split_build_module(target_name)
        return '{} (in {})'.format(target_name,
                                   conf.get_build_file_path(build_module))

    def format_unresolved(seed):
        if seed not in seed_refs:
            return seed
        seed_ref = seed_refs[seed]
        reasons = []
        if seed_ref.on_cli:
            reasons.append('seen on command line')
        if seed_ref.from_default:
            reasons.append('specified as default target in {}'
                           .format(conf.get_project_build_file))
        if seed_ref.dep_of:
            reasons.append(
                'dependency of ' +
                ', '.join(format_target(target_name)
                          for target_name in sorted(seed_ref.dep_of)))
        if seed_ref.buildenv_of:
            reasons.append(
                'buildenv of ' +
                ', '.join(format_target(target_name)
                          for target_name in sorted(seed_ref.buildenv_of)))

        return '{} (possible misspelling of {}) - {}'.format(
            seed, difflib.get_close_matches(
                seed, build_context.targets.keys()), ', '.join(reasons))

    unresolved_str = '\n'.join(format_unresolved(target_name)
                               for target_name in sorted(unknown_seeds))
    num_target_str = '{} target'.format(len(unknown_seeds))
    if len(unknown_seeds) > 1:
        num_target_str += 's'
    raise ValueError('Could not resolve {}:\n{}'
                     .format(num_target_str, unresolved_str))


def populate_targets_graph(build_context, conf: Config):
    # Process project root build file
    process_build_file(conf.get_project_build_file(), build_context, conf)
    targets_to_prune = set(build_context.targets.keys())
    seed_refs = defaultdict(SeedRef)
    if conf.targets:
        logger.debug('targets: {}', conf.targets)
        # TODO(itamar): Figure out how to support a target selector that is a
        #   parent directory which isn't a build module, but contains build
        #   modules (e.g., `ybt tree yapi` from the `dag` test root).
        seeds = parse_target_selectors(conf.targets, conf)
        logger.debug('seeds: {}', seeds)
        for cli_seed in seeds:
            seed_refs[cli_seed].on_cli = True
    else:
        default_target = ':{}'.format(conf.default_target_name)
        logger.info('searching for default target {}', default_target)
        if default_target not in build_context.targets:
            raise RuntimeError(
                'No default target found in {}, and no target selector '
                'specified'.format(conf.get_project_build_file()))
        seeds = [default_target]
        seed_refs[default_target].from_default = True

    def extend_seeds(target_name):
        target = build_context.targets[target_name]
        seeds.extend(target.deps)
        for dep in target.deps:
            seed_refs[dep].dep_of.add(target_name)
        if target.buildenv:
            seeds.append(target.buildenv)
            seed_refs[target.buildenv].buildenv_of.add(target_name)

    # Crawl rest of project from seeds
    seeds_used_for_extending = set()
    unknown_seeds = set()
    for seed in seeds:
        if seed in build_context.targets:
            #
            if seed in targets_to_prune:
                targets_to_prune.remove(seed)
            if seed not in seeds_used_for_extending:
                # Avoid infinite loop in case of cyclic dependencies
                extend_seeds(seed)
                seeds_used_for_extending.add(seed)
        else:
            if seed == '**:*':
                # Adding all build modules under current working directory as
                # seeds
                seeds.extend(generate_all_targets(conf))
                continue
            build_module, target_name = split(seed)
            process_build_file(conf.get_build_file_path(build_module),
                               build_context, conf)
            # Parsed build file with this seed target - add its dependencies as
            # seeds
            if target_name == '*':
                # It's a wildcard - add all targets from build module
                # (and skip adding to targets_to_prune altogether)
                for module_target in (
                        build_context.targets_by_module[build_module]):
                    extend_seeds(module_target)
            else:
                if seed not in build_context.targets:
                    unknown_seeds.add(seed)
                    continue
                for module_target in (
                        build_context.targets_by_module[build_module]):
                    targets_to_prune.add(module_target)
                targets_to_prune.remove(seed)
                extend_seeds(seed)
        # TODO(itamar): Write tests that pruning is *ALWAYS* correct!
        # e.g., not pruning things it shouldn't (like when targets are in prune
        # list when loaded initially, but should be removed later because a
        # target loaded later required them).

    if unknown_seeds:
        raise_unresolved_targets(build_context, conf, unknown_seeds, seed_refs)

    # Pruning, after parsing is done
    # (first, adding targets that are tagged as "prune-me" to prune list)
    targets_to_prune.update(
        target_name
        for target_name, target in build_context.targets.items()
        if 'prune-me' in target.tags)
    for target_name in targets_to_prune:
        build_context.remove_target(target_name)

    # fill in an actual target graph using the dependencies info
    # and assert that it has no cycles
    build_target_dep_graph(build_context, conf)
    if not dag.is_directed_acyclic_graph(build_context.target_graph):
        cycles = '\n'.join(
            ' -> '.join(cycle)
            for cycle in networkx.simple_cycles(build_context.target_graph))
        raise RuntimeError('Detected cycles in build graph!\n' + cycles)

    # go over the graph and assert policies
    violations = []
    if conf.policies:
        for target in build_context.targets.values():
            for policy in conf.policies:
                err = policy(build_context, target)
                if err:
                    violations.append(
                        '- Target {} violates {} policy: {}'.format(
                            target.name, policy.__name__, err))
    if violations:
        raise RuntimeError('Policy violations:\n' + '\n'.join(violations))

    logger.info('Finished parsing build graph with {} nodes and {} edges',
                build_context.target_graph.order(),
                build_context.target_graph.size())


def top_rev_sort_subgraph_stable(graph):
    """
    This is a  topological sorting algorithm stable for all subgraphs
    L <- Empty queue that will contain the sorted elements
    S <- Set of all nodes with no outcoming edges

    while S is non-empty do
    remove a node n from S
    add n to tail of L
    for each node m with an edge e from n to m do
        remove edge e from the graph
        if m has no other outcoming edges then
        insert m into S

    if graph has edges then
    return error     (graph has at least one cycle)
    else
    return L     (a topologically sorted order)
    """
    if not graph.is_directed():
        raise networkx.NetworkXError(
            'Topological sort not defined on undirected graphs.')
    graph = graph.copy()
    visited = dict.fromkeys(graph.nodes(), False)
    while True:
        set_s = []
        for node in graph.nodes():
            if not visited[node]:
                if graph.out_degree(node) == 0:
                    visited[node] = True
                    set_s.append(node)
        set_s.sort()
        if not set_s:
            break
        for node in set_s:
            yield node
            graph.remove_node(node)
    if graph:
        raise networkx.NetworkXUnfeasible('Graph contains a cycle.')


def topological_sort(graph: networkx.DiGraph):
    yield from top_rev_sort_subgraph_stable(graph)


def get_descendants(graph: networkx.DiGraph, source):
    """Return all nodes reachable from `source` in `graph`."""
    return dag.descendants(graph, source)


def get_ancestors(graph: networkx.DiGraph, child):
    """Return all nodes that have 'child' as a descendant in 'graph'"""
    return dag.ancestors(graph, child)


def get_graph_roots(graph):
    for id_deg in graph.in_degree():
        if id_deg[1] == 0:
            yield id_deg[0]


def cut_from_graph(graph, name):
    graph = graph.copy()
    cut_s = {name}
    cut_s.update(dag.descendants(graph, name))
    nodes = list(graph.nodes())
    for node in nodes:
        if node not in cut_s:
            graph.remove_node(node)
    return graph



# -*- coding: utf-8 -*-

# Copyright 2021 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Gives information about the target
~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

import json

from .config import Config
from .target_utils import parse_target_selectors


def get_target_info_json(conf: Config, build_context):
    targets = parse_target_selectors(conf.targets, conf)
    targets_info = {}
    for target_name in targets:
        target = build_context.targets[target_name]
        info = {
            'workspace': build_context.get_workspace(target.builder_name,
                                                     target_name)
        }
        if 'image_caching_behavior' in target.props:
            if 'remote_image_name' in target.props.image_caching_behavior:
                info['remote_image_name'] = \
                    target.props.image_caching_behavior['remote_image_name']
            if 'remote_image_tag' in target.props.image_caching_behavior:
                info['remote_image_tag'] = \
                    target.props.image_caching_behavior['remote_image_tag']
        targets_info[target_name] = info
    return json.dumps(targets_info)



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

:author: Zohar Rimon
"""


import os
from os.path import isdir, isfile, join
import shutil
from subprocess import PIPE

import pytest

from conftest import reset_parser
from .. import cli, extend
from . import proto
from ..buildcontext import BuildContext
from ..graph import populate_targets_graph
from ..utils import yprint


PROTO_GEN_DIR = 'ybtproto'  # as defined in tests/proto/YSettings


def clear_output():
    try:
        shutil.rmtree('build')
    except FileNotFoundError:
        pass


@pytest.mark.slow
@pytest.mark.usefixtures('in_proto_project')
def test_proto_builder(basic_conf):
    clear_output()
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['app:hello-proto']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph()
    assert isdir('build')
    assert isdir(join('build', 'gen'))
    assert isdir(join('build', 'gen', PROTO_GEN_DIR))
    assert isfile(join('build', 'gen', PROTO_GEN_DIR, '__init__.py'))
    assert isdir(join('build', 'gen', PROTO_GEN_DIR, 'app'))
    assert isfile(join('build', 'gen', PROTO_GEN_DIR, 'app', '__init__.py'))
    for exp_gen_fname in [
        'hello.pb.cc',
        'hello.pb.h',
        'hello_pb2.py'
    ]:
        assert isfile(
          join('build', 'gen', PROTO_GEN_DIR, 'app', exp_gen_fname))
    clear_output()


@pytest.mark.slow
@pytest.mark.usefixtures('in_proto_project')
def test_proto_cpp_prog(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['app:hello-prog']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph()
    work_dir = build_context.conf.host_to_buildenv_path(
        build_context.get_workspace('CppProg', 'app:hello-prog'))
    result = build_context.run_in_buildenv(
        ':proto-builder', [join(work_dir, 'app', 'hello-prog')],
        stdout=PIPE, stderr=PIPE)
    assert 0 == result.returncode
    assert b'Hello, World!' == result.stdout


@pytest.mark.slow
@pytest.mark.usefixtures('in_proto_project')
def test_proto_collector(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = ['app:hello1-collector']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph()
    assert_all_proto_files_exist()


@pytest.mark.slow
@pytest.mark.usefixtures('in_proto_project')
def test_proto_collector_build_from_cache():
    # Create cache
    reset_parser()
    conf = cli.init_and_get_conf(['--non-interactive', 'build'])
    extend.Plugin.load_plugins(conf)
    build_context = BuildContext(conf)
    conf.targets = ['app:hello1-collector']
    populate_targets_graph(build_context, conf)
    build_context.build_graph()

    # Build from cache
    shutil.rmtree(join('yabtwork', 'flavor__all__'))
    build_context2 = BuildContext(conf)
    populate_targets_graph(build_context2, conf)
    build_context2.build_graph()
    assert_all_proto_files_exist()


def assert_all_proto_files_exist():
    assert isdir('yabtwork')
    assert isdir(join('yabtwork', 'flavor__all__'))
    assert isdir(join('yabtwork', 'flavor__all__', 'ProtoCollector'))
    assert isdir(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                      'app_hello1-collector'))
    assert isdir(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                      'app_hello1-collector', PROTO_GEN_DIR))
    assert isdir(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                      'app_hello1-collector', PROTO_GEN_DIR, 'app'))
    assert isdir(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                      'app_hello1-collector', PROTO_GEN_DIR, 'app', 'hello1'))
    assert isfile(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                       'app_hello1-collector', PROTO_GEN_DIR, 'app', 'hello1',
                       'hello1.proto'))
    assert isdir(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                      'app_hello1-collector', PROTO_GEN_DIR, 'app', 'hello2'))
    assert isfile(join('yabtwork', 'flavor__all__', 'ProtoCollector',
                       'app_hello1-collector', PROTO_GEN_DIR, 'app', 'hello2',
                       'hello2.proto'))



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Compatability module
~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


# os.scandir, performant os.walk
try:
    # Python 3.5+, use builtin implementation
    # https://docs.python.org/3/library/os.html#os.scandir
    from os import scandir, walk
except ImportError:
    # pre-Python 3.5, use third party package
    # https://pypi.python.org/pypi/scandir
    from scandir import scandir, walk



# -*- coding: utf-8 -*-

# Copyright 2020 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Go Builder Tests

:author: Itamar Ostricher
"""


from subprocess import check_output

import pytest

from ..buildcontext import BuildContext
from ..graph import populate_targets_graph


@pytest.mark.slow
@pytest.mark.usefixtures('in_golang_project')
def test_golang_builder(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'hello:hello-app'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)
    hello_out = check_output(
        ['docker', 'run', '--rm', build_context.targets[target_name].image_id,
         '-who', 'app'])
    expected = br"""
  _              _   _
 | |__     ___  | | | |   ___       __ _   _ __    _ __
 | '_ \   / _ \ | | | |  / _ \     / _` | | '_ \  | '_ \
 | | | | |  __/ | | | | | (_) |   | (_| | | |_) | | |_) |
 |_| |_|  \___| |_| |_|  \___/     \__,_| | .__/  | .__/
                                          |_|     |_|
"""
    assert expected == (b'\n' + hello_out)


@pytest.mark.slow
@pytest.mark.usefixtures('in_golang_project')
def test_golang_test(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'hello_lib:utils_test'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)


@pytest.mark.slow
@pytest.mark.usefixtures('in_golang_project')
def test_golang_test_multisrc(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'goodbye_lib:goodbye_utils_test'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)


@pytest.mark.slow
@pytest.mark.usefixtures('in_golang_project')
def test_golang_test_mainfuncs(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'goodbye:goodbye-test'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)


@pytest.mark.slow
@pytest.mark.usefixtures('in_proto_project')
def test_golang_builder_proto(basic_conf):
    build_context = BuildContext(basic_conf)
    target_name = 'app:hello-go-app'
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)
    hello_out = check_output(
        ['docker', 'run', '--rm', build_context.targets[target_name].image_id,
         '-who', 'boomer'])
    assert hello_out == b'message: "hello boomer"\n\n'



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt source control subsystem
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from abc import abstractmethod, ABCMeta
import pkg_resources

from .logging import make_logger


logger = make_logger(__name__)


class SourceControl(metaclass=ABCMeta):
    """Source Control interface Abstract Base Class"""

    @abstractmethod
    def __init__(self, conf):
        """Initialize SourceControl instance with config instance.

        :param conf: A yabt.config.Config object.
        """

    @abstractmethod
    def get_revision(self) -> str:
        """Return revision string of active repo tip / head."""


class ScmManager:
    """Source Control subsystem manager.

    Holds a dictionary of registered concrete SCM providers.
    """

    providers = {}

    @classmethod
    def get_provider(cls, scm_name: str, conf) -> SourceControl:
        """Load and return named SCM provider instance.

        :param conf: A yabt.config.Config object used to initialize the SCM
                        provider instance.

        :raises KeyError: If no SCM provider with name `scm_name` registered.
        """
        for entry_point in pkg_resources.iter_entry_points('yabt.scm',
                                                           scm_name):
            entry_point.load()
            logger.debug('Loaded SCM provider {0.name} from {0.module_name} '
                         '(dist {0.dist})', entry_point)
        logger.debug('Loaded {} SCM providers', len(cls.providers))
        if scm_name not in cls.providers:
            raise KeyError('Unknown SCM identifier {}'.format(scm_name))
        return cls.providers[scm_name](conf)

    def __init__(self, conf):
        """
        :param conf: A yabt.config.Config object.
        """
        self.conf = conf


def register_scm_provider(scm_name: str):
    """Return a decorator for registering a SCM provider named `scm_name`."""

    def register_decorator(scm_class: SourceControl):
        """Decorator for registering SCM provider."""
        if scm_name in ScmManager.providers:
            raise KeyError('{} already registered!'.format(scm_name))
        ScmManager.providers[scm_name] = scm_class
        SourceControl.register(scm_class)
        logger.debug('Registered {0} SCM from {1.__module__}.{1.__name__}',
                     scm_name, scm_class)
        return scm_class
    return register_decorator


@register_scm_provider('none')
class NoSCM:
    """Default trivial (AKA non-existent) concrete SCM implementation."""

    def __init__(self, unused_conf):
        pass

    def get_revision(self):
        raise NotImplementedError(
            'SCM provider not specified or not configured correctly')



from subprocess import check_output
import pytest

from yabt.buildcontext import BuildContext
from yabt.graph import populate_targets_graph


@pytest.mark.slow
@pytest.mark.usefixtures('in_custom_installer_project')
def test_custom_installer_args(basic_conf):
    build_context = BuildContext(basic_conf)
    basic_conf.targets = [':w00t']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)
    w00t_content = check_output(
        ['docker', 'run', '--rm', '--entrypoint', 'cat',
         build_context.targets[':w00t'].image_id, '/var/opt/w00t'])
    assert w00t_content == b'hello there\nhow are you doing\n'



# -*- coding: utf-8 -*-

# Copyright 2019 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Creates the dot graph
~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

from .caching import get_prebuilt_targets
from .config import Config


TARGETS_COLORS = {'AptPackage': 'brown4',
                  'CppGTest': 'deepskyblue4',
                  'CppLib': 'blue',
                  'CppProg': 'cornflowerblue',
                  'CustomInstaller': 'brown',
                  'Proto': 'green',
                  'Python': 'red',
                  'PythonPackage': 'purple',
                  'PythonTest': 'pink'
                  }


def get_not_buildenv_targets(build_context):
    roots = [n for n, d in build_context.target_graph.in_degree() if d == 0]
    buildenvs = set(
        target.buildenv for target in build_context.targets.values()
        if target.buildenv is not None)
    visited = set()
    to_do = []
    for root in roots:
        if root not in buildenvs:
            visited.add(root)
            to_do.append(root)

    while to_do:
        node = to_do.pop(-1)
        for successor in build_context.target_graph.successors(node):
            if successor not in visited and successor not in buildenvs:
                visited.add(successor)
                to_do.append(successor)
    return visited


def write_dot(build_context, conf: Config, out_f):
    """Write build graph in dot format to `out_f` file-like object."""
    not_buildenv_targets = get_not_buildenv_targets(build_context)
    prebuilt_targets = get_prebuilt_targets(build_context)
    out_f.write('strict digraph  {\n')
    for node in build_context.target_graph.nodes:
        if conf.show_buildenv_deps or node in not_buildenv_targets:
            cached = node in prebuilt_targets
            fillcolor = 'fillcolor="grey",style=filled' if cached else ''
            color = TARGETS_COLORS.get(
                build_context.targets[node].builder_name, 'black')
            out_f.write('  "{}" [color="{}",{}];\n'.format(node, color,
                                                           fillcolor))
    out_f.writelines('  "{}" -> "{}";\n'.format(u, v)
                     for u, v in build_context.target_graph.edges
                     if conf.show_buildenv_deps or
                     (u in not_buildenv_targets and v in not_buildenv_targets))
    out_f.write('}\n\n')



# -*- coding: utf-8 -*-

# Copyright 2021 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
cpp caching tests
~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""

from os import path
import os
from subprocess import check_output

import pytest
import shutil

from .. import cli
from ..buildcontext import BuildContext
from ..graph import populate_targets_graph

PROJECT_DIR = path.join(path.dirname(path.abspath(__file__)), '..', '..',
                        'tests', 'cpp_caching')
OP_OBJ_FILE = path.join('yabtwork', 'release_flavor', 'CppLib',
                        '_op_user_lib', 'src', 'op_user_lib.o')


def build_main_app(target_name):
    basic_conf = cli.init_and_get_conf(
        ['--non-interactive', '--continue-after-fail', '--scm-provider',
         'none', 'build'])
    build_context = BuildContext(basic_conf)
    basic_conf.targets = [target_name]
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph()


def build_test():
    basic_conf = cli.init_and_get_conf(
        ['--non-interactive', '--continue-after-fail', '--scm-provider',
         'none', 'build'])
    build_context = BuildContext(basic_conf)
    basic_conf.targets = [':op_user_test']
    populate_targets_graph(build_context, basic_conf)
    build_context.build_graph(run_tests=True)


@pytest.mark.slow
def test_caching_prog(tmp_dir):
    shutil.copytree(PROJECT_DIR, 'cpp_caching')
    os.chdir('cpp_caching')

    build_main_app(':main-app')
    op_obj_timestamp = path.getmtime(OP_OBJ_FILE)
    assert check_output(['docker', 'run', 'main-app:latest']) == b'12'

    with open('binary_operation.cc', 'r') as f:
        binary_operation_code = f.read()
    with open('binary_operation.cc', 'w') as f:
        f.write(binary_operation_code.replace('+', '*'))

    build_main_app(':main-app')
    assert op_obj_timestamp == path.getmtime(OP_OBJ_FILE)
    assert check_output(['docker', 'run', 'main-app:latest']) == b'20'


@pytest.mark.slow
def test_caching_gtest(tmp_dir):
    shutil.copytree(PROJECT_DIR, 'cpp_caching')
    os.chdir('cpp_caching')

    build_test()
    op_obj_timestamp = path.getmtime(OP_OBJ_FILE)

    with open('binary_operation.cc', 'r') as f:
        binary_operation_code = f.read()
    with open('binary_operation.cc', 'w') as f:
        f.write(binary_operation_code.replace('+', '*'))

    with pytest.raises(SystemExit):
        build_test()
    assert op_obj_timestamp == path.getmtime(OP_OBJ_FILE)


@pytest.mark.slow
def test_caching_far_change(tmp_dir):
    shutil.copytree(PROJECT_DIR, 'cpp_caching')
    os.chdir('cpp_caching')

    build_main_app(':main_far_change-app')
    op_obj_timestamp = path.getmtime(OP_OBJ_FILE)
    assert check_output(['docker', 'run', 'main_far_change-app:latest']) \
           == b'12'

    with open('binary_operation.cc', 'r') as f:
        binary_operation_code = f.read()
    with open('binary_operation.cc', 'w') as f:
        f.write(binary_operation_code.replace('+', '*'))

    build_main_app(':main_far_change-app')
    assert op_obj_timestamp == path.getmtime(OP_OBJ_FILE)
    assert check_output(['docker', 'run', 'main_far_change-app:latest']) \
           == b'20'



# -*- coding: utf-8 -*-

# Copyright 2021 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt target info tests
~~~~~~~~~~~~~~~~~~~~~~~

:author: Dana Shamir
"""
import json
import os
import pytest

from .buildcontext import BuildContext
from .graph import populate_targets_graph
from .target_info import get_target_info_json


@pytest.mark.usefixtures('in_caching_project')
def test_target_info(basic_conf):
    basic_conf.targets = [':builder-base', ':builder']
    build_context = BuildContext(basic_conf)
    populate_targets_graph(build_context, basic_conf)
    result = get_target_info_json(basic_conf, build_context)
    expected = {
        ':builder-base':
            {'workspace': os.path.join(
                os.getcwd(), 'yabtwork', 'flavor__all__', 'DockerImage',
                '_builder-base'),
             'remote_image_name': 'itamarost/builder-base-test',
             'remote_image_tag': 'v1'},
        ':builder':
            {'workspace': os.path.join(
                os.getcwd(), 'yabtwork', 'flavor__all__', 'DockerImage',
                '_builder')}
    }
    assert json.loads(result) == expected



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt cli module
~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""

import json
from importlib.machinery import SourceFileLoader
import os

import argcomplete
import colorama
import configargparse

from ostrich.utils.collections import listify

from .config import BUILD_PROJ_FILE, Config, YCONFIG_FILE, YSETTINGS_FILE
from .utils import search_for_parent_dir


PARSER = None
KNOWN_LOG_LEVELS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
LOG_LEVELS_CHOICES = KNOWN_LOG_LEVELS + [level.lower()
                                         for level in KNOWN_LOG_LEVELS]


def make_parser(project_config_file: str) -> configargparse.ArgumentParser:
    """Return the argument parser.

    :param project_config_file: Absolute path to project-specific config file.

    If cached parser already exists - return it immediately.
    Otherwise, initialize a new `ConfigArgParser` that is able to take default
    values from a hierarchy of config files and environment variables, as well
    as standard ArgParse command-line parsing behavior.

    We take default values from configuration files:

    - System-wide (see code for location)
    - User-level overrides (see code for location, hopefully under home dir)
    - If a project-specific config file is available, it will override both
      of the above.

    Environment variables will override all configuration files.
    For an option `--foo-bar`, if an environment variable named `YBT_FOO_VAR`
    exists, the option value will be taken from there.

    Of course, options specified directly on the command-line always win.
    """
    global PARSER  # pylint: disable=global-statement
    if PARSER is None:
        config_files = ['/etc/yabt.conf', '~/.yconfig']
        if project_config_file:
            config_files.append(project_config_file)
        PARSER = configargparse.getArgumentParser(
            # Support loading default values from system-wide or
            # user-specific config files (user-level overrides system-wide)
            default_config_files=config_files,
            # Show default values in help message
            formatter_class=configargparse.DefaultsFormatter,
            auto_env_var_prefix='ybt_',
            args_for_setting_config_path=['--config'],
            args_for_writing_out_config_file=['--write-out-config-file'])
        # PARSER.add('--config', is_config_file=True, help='Config file path')
        PARSER.add('--artifacts-metadata-file',
                   help='Output file to write artifacts metadata to')
        PARSER.add('--continue-after-fail', default=False, action='store_true',
                   help='If a target fails continue independent targets')
        PARSER.add('--bin-output-dir', default='ybt_bin')
        PARSER.add('--build-file-name', default='YBuild')
        PARSER.add('--build-base-images', action='store_true')
        PARSER.add('--builders-workspace-dir', default='yabtwork')
        PARSER.add('--default-target-name', default='@default')
        PARSER.add('--docker-pull-cmd', default='docker pull',
                   help='Command to use for pulling images from registries')
        PARSER.add('--docker-push-cmd', default='docker push',
                   help='Command to use for pushing images to registries')
        PARSER.add('--docker-volume',
                   help='Use the specified docker volume as buildenv /project')
        PARSER.add('-f', '--flavor', help='Choose build flavor (AKA profile)')
        PARSER.add('--force-pull', action='store_true')
        PARSER.add('-j', '--jobs', type=int, default=1)
        # TODO(itamar): support auto-detection of interactivity-mode
        PARSER.add('--non-interactive', action='store_true')
        PARSER.add('--offline', action='store_true')
        PARSER.add('--output-dot-file', default=None,
                   help='Output file for dot graph (default: stdin)')
        # TODO(itamar): this flag should come from the builder, not from here
        PARSER.add('--push', action='store_true')
        PARSER.add('--runtime-params', type=json.loads,
                   help='Params to pass to the docker run command in json')
        PARSER.add('--scm-provider')
        PARSER.add('--no-build-cache', action='store_true',
                   help='Disable YBT build cache')
        PARSER.add('--no-docker-cache', action='store_true',
                   help='Disable YBT Docker cache')
        PARSER.add('--no-policies', action='store_true')
        PARSER.add('--no-test-cache', action='store_true',
                   help='Disable YBT test cache')
        PARSER.add('--test-attempts', type=int, default=1)
        PARSER.add('-v', '--verbose', action='store_true',
                   help='More verbose output to STDOUT')
        PARSER.add('--with-tini-entrypoint', action='store_true')
        # Logging flags
        PARSER.add('--logtostderr', action='store_true',
                   help='Whether to log to STDERR')
        PARSER.add('--logtostdout', action='store_true',
                   help='Whether to log to STDOUT')
        PARSER.add('--loglevel', default='INFO', choices=LOG_LEVELS_CHOICES,
                   help='Log level threshold')
        PARSER.add('--show-buildenv-deps', type=bool, default=False,
                   help='When running dot, if set to True then the buildenv '
                        'targets are printed to the graph too')
        PARSER.add('--download-from-global-cache', default=False,
                   action='store_true',
                   help='download from global cache targets that are not '
                        'cached locally')
        PARSER.add('--upload-to-global-cache', default=False,
                   action='store_true',
                   help='upload to global cache targets that were built')
        PARSER.add('--download-tests-from-global-cache', default=False,
                   action='store_true',
                   help='download from global cache tests that are not '
                        'cached locally')
        PARSER.add('--upload-tests-to-global-cache', default=False,
                   action='store_true',
                   help='upload to global cache tests that were run')
        PARSER.add('cmd', choices=['build', 'dot', 'test', 'tree', 'version',
                                   'info'])
        PARSER.add('targets', nargs='*')
    return PARSER


def find_project_config_file(project_root: str) -> str:
    """Return absolute path to project-specific config file, if it exists.

    :param project_root: Absolute path to project root directory.

    A project config file is a file named `YCONFIG_FILE` found at the top
    level of the project root dir.

    Return `None` if project root dir is not specified,
    or if no such file is found.
    """
    if project_root:
        project_config_file = os.path.join(project_root, YCONFIG_FILE)
        if os.path.isfile(project_config_file):
            return project_config_file


def get_user_settings_module(project_root: str):
    """Return project-specific user settings module, if it exists.

    :param project_root: Absolute path to project root directory.

    A project settings file is a file named `YSETTINGS_FILE` found at the top
    level of the project root dir.

    Return `None` if project root dir is not specified,
    or if no such file is found.

    Raise an exception if a file is found, but not importable.

    The YSettings file can define 2 special module-level functions that
    interact with the YABT CLI & config system:
    1. `extend_cli`, if defined, takes the YABT `parser` object and may extend
       it, to add custom command-line flags for the project.
       (careful not to collide with YABT flags...)
    2. `extend_config`, if defined, takes the YABT `config` object and the
       parsed `args` object (returned by the the parser), and may extend the
       config - should be used to reflect custom project CLI flags in the
       config object.

    Beyond that, the settings module is available in YBuild's under
    `conf.settings` (except for the 2 special fucntions that are removed).
    """
    if project_root:
        project_settings_file = os.path.join(project_root, YSETTINGS_FILE)
        if os.path.isfile(project_settings_file):
            settings_loader = SourceFileLoader(
                'settings', project_settings_file)
            return settings_loader.load_module()


def call_user_func(settings_module, func_name, *args, **kwargs):
    """Call a user-supplied settings function and clean it up afterwards.

    settings_module may be None, or the function may not exist.
    If the function exists, it is called with the specified *args and **kwargs,
    and the result is returned.
    """
    if settings_module:
        if hasattr(settings_module, func_name):
            func = getattr(settings_module, func_name)
            try:
                return func(*args, **kwargs)
            finally:
                # cleanup user function
                delattr(settings_module, func_name)


def get_build_flavor(settings_module, args):
    """Update the flavor arg based on the settings API"""
    known_flavors = listify(call_user_func(settings_module, 'known_flavors'))
    if args.flavor:
        if args.flavor not in known_flavors:
            raise ValueError('Unknown build flavor: {}'.format(args.flavor))
    else:
        args.flavor = call_user_func(settings_module, 'default_flavor')
        if args.flavor and args.flavor not in known_flavors:
            raise ValueError(
                'Unknown default build flavor: {}'.format(args.flavor))


def init_and_get_conf(argv: list=None) -> Config:
    """Initialize a YABT CLI environment and return a Config instance.

    :param argv: Manual override of command-line params to parse (for tests).
    """
    colorama.init()
    work_dir = os.path.abspath(os.curdir)
    project_root = search_for_parent_dir(work_dir,
                                         with_files=set([BUILD_PROJ_FILE]))
    parser = make_parser(find_project_config_file(project_root))
    settings_module = get_user_settings_module(project_root)
    call_user_func(settings_module, 'extend_cli', parser)
    argcomplete.autocomplete(parser)
    args = parser.parse(argv)
    get_build_flavor(settings_module, args)
    config = Config(args, project_root, work_dir, settings_module)
    config.common_conf = call_user_func(
        config.settings, 'get_common_config', config, args)
    config.flavor_conf = call_user_func(
        config.settings, 'get_flavored_config', config, args)
    call_user_func(config.settings, 'extend_config', config, args)
    if not args.no_policies:
        config.policies = listify(call_user_func(
            config.settings, 'get_policies', config))
    return config



# -*- coding: utf-8 -*-

# Copyright 2018 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Caching module
~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher

TODO: implement also distributed caching
will require keeping track of which machine produced what, so it is possible
to rerun "cached" tests on machines with different hardware (for example).
"""
import itertools
import json
import os
from functools import partial
from os import makedirs
from os.path import isdir, isfile, join, relpath, split, dirname
import shutil
from time import time

from ostrich.utils.text import get_safe_path

from .artifact import ArtifactType as AT
from .config import Config
from .docker import get_image_name, handle_build_cache, tag_docker_image
from .graph import get_descendants
from .logging import make_logger
from .target_utils import ImageCachingBehavior, Target
from .utils import hash_tree, rmnode, rmtree


logger = make_logger(__name__)

_NO_CACHE_TYPES = frozenset((AT.app, AT.docker_image))
MAX_FAILS_FROM_GLOBAL = 5


class CachedDescendants(dict):
    """Utility dict-like class for holding mappings from target-name to set of
       descendants of that target.
    """

    def __init__(self, target_graph):
        self._target_graph = target_graph
        super().__init__()

    def get(self, key):
        """Return set of descendants of node named `key` in `target_graph`.

        Returns from cached dict if exists, otherwise compute over the graph
        and cache results in the dict.
        """
        if key not in self:
            self[key] = set(get_descendants(self._target_graph, key))
        return self[key]


def get_prebuilt_targets(build_context):
    """Return set of target names that are contained within cached base images

    These targets may be considered "pre-built", and skipped during build.
    """
    logger.info('Scanning for cached base images')
    # deps that are part of cached based images
    contained_deps = set()
    # deps that are needed by images that are going to be built,
    # but are not part of their base images
    required_deps = set()
    # mapping from target name to set of all its deps (descendants)
    cached_descendants = CachedDescendants(build_context.target_graph)

    for target_name, target in build_context.targets.items():
        if 'image_caching_behavior' not in target.props:
            continue
        image_name = get_image_name(target)
        image_tag = target.props.image_tag
        icb = ImageCachingBehavior(image_name, image_tag,
                                   target.props.image_caching_behavior)
        target.image_id = handle_build_cache(build_context.conf, image_name,
                                             image_tag, icb)
        if target.image_id:
            # mark deps of cached base image as "contained"
            image_deps = cached_descendants.get(target_name)
            contained_deps.update(image_deps)
            contained_deps.add(target.name)
        else:
            # mark deps of image that is going to be built
            # (and are not deps of its base image) as "required"
            image_deps = cached_descendants.get(target_name)
            base_image_deps = cached_descendants.get(target.props.base_image)
            required_deps.update(image_deps - base_image_deps)
    return contained_deps - required_deps


def write_summary(summary: dict, cache_dir: str):
    """Write the `summary` JSON to `cache_dir`.

    Updated the accessed timestamp to now before writing.
    """
    # update the summary last-accessed timestamp
    summary['accessed'] = time()
    with open(join(cache_dir, 'summary.json'), 'w') as summary_file:
        summary_file.write(json.dumps(summary, indent=4, sort_keys=True))


def load_target_from_global_cache(target: Target, build_context) -> bool:
    target_hash = target.hash(build_context)
    if not build_context.global_cache.has_cache(target_hash):
        return False
    cache_dir = build_context.conf.get_cache_dir(target, build_context)
    makedirs(cache_dir, exist_ok=True)
    if not build_context.global_cache.download_summary(
            target_hash, join(cache_dir, 'summary.json')):
        return False
    if not build_context.global_cache.download_artifacts_meta(
            target_hash, join(cache_dir, 'artifacts.json')):
        return False
    with open(join(cache_dir, 'artifacts.json'), 'r') as artifacts_meta_file:
        artifacts_desc = json.load(artifacts_meta_file)
    makedirs(build_context.conf.get_artifacts_cache_dir(), exist_ok=True)
    if not build_context.global_cache.download_artifacts(
        get_artifacts_hashes(artifacts_desc),
            build_context.conf.get_artifacts_cache_dir()):
        return False
    if build_context.conf.download_tests_from_global_cache:
        build_context.global_cache.download_test_cache(
            target_hash, join(cache_dir, 'tested.json'))
    return True


def get_artifacts_hashes(artifact_desc):
    """
    Returns a dict containing all the artifacts hashes as keys and the
    permissions required for the artifact.
    """
    artifacts_hashes = {}
    for type_name, artifact_list in artifact_desc.items():
        artifact_type = getattr(AT, type_name)
        for artifact in artifact_list:
            if artifact_type not in _NO_CACHE_TYPES and 'hash' in artifact \
                    and artifact['hash'] is not None:
                artifacts_hashes[artifact['hash']] = artifact['permissions']
    return artifacts_hashes


def load_target_from_cache(target: Target, build_context) -> (bool, bool):
    """Load `target` from build cache, restoring cached artifacts & summary.
       Return (build_cached, test_cached) tuple.

    `build_cached` is True if target restored successfully.
    `test_cached` is True if build is cached and test_time metadata is valid.
    """
    cache_dir = build_context.conf.get_cache_dir(target, build_context)
    if not isdir(cache_dir):
        logger.debug('No cache dir found for target {}', target.name)
        has_global_cache = False
        if build_context.global_cache and \
                build_context.conf.download_from_global_cache:
            logger.info('Trying to load target {} from global cache',
                        target.name)
            has_global_cache = try_use_global_cache(
                build_context, partial(load_target_from_global_cache, target,
                                       build_context),
                'an error occurred while trying to download target {} from '
                'global cache'.format(target.name))
        if not has_global_cache:
            try:
                shutil.rmtree(cache_dir)
            except FileNotFoundError:
                pass
            return False, False
    try:
        # read summary file and restore relevant fields into target
        with open(join(cache_dir, 'summary.json'), 'r') as summary_file:
            summary = json.loads(summary_file.read())
        for field in ('build_time', 'test_time', 'created', 'accessed'):
            target.summary[field] = summary.get(field)
        # compare artifacts hash
        if (hash_tree(join(cache_dir, 'artifacts.json')) !=
                summary.get('artifacts_hash', 'no hash')):
            return False, False
        # read cached artifacts metadata
        with open(join(cache_dir, 'artifacts.json'), 'r') \
                as artifacts_meta_file:
            artifact_desc = json.loads(artifacts_meta_file.read())
        # restore all artifacts
        for type_name, artifact_list in artifact_desc.items():
            artifact_type = getattr(AT, type_name)
            for artifact in artifact_list:
                # restore artifact to its expected src path
                if artifact_type not in _NO_CACHE_TYPES:
                    if not restore_artifact(artifact['src'], artifact['hash'],
                                            build_context.conf):
                        target.artifacts.reset()
                        return False, False
                if artifact_type in (AT.docker_image,):
                    # "restore" docker image from local registry
                    image_id = artifact['src']
                    image_full_name = artifact['dst']
                    try:
                        tag_docker_image(image_id, image_full_name)
                    except:
                        logger.debug('Docker image with ID {} not found '
                                     'locally', image_id)
                        target.artifacts.reset()
                        return False, False
                    target.image_id = image_id
                target.artifacts.add(
                    artifact_type, artifact['src'], artifact['dst'])
        write_summary(summary, cache_dir)
    except (json.decoder.JSONDecodeError, FileNotFoundError) as ex:
        logger.warning("Got {} when trying to read cache of "
                       "target {}. There is probably a corrupted file. "
                       "Deleting the cache and not using it.",
                       type(ex).__name__, target.name)
        try:
            shutil.rmtree(cache_dir)
        except FileNotFoundError:
            pass
        return False, False

    # check that the testing cache exists.
    if not isfile(join(cache_dir, 'tested.json')):
        logger.debug('No testing cache found for target {}', target.name)
        return True, False

    # TODO(Dana) - make PythonTest run without mounting the project directory
    # so that the dependencies will be hermetic, and so is the cache.
    # Then we will be able to cache python tests.
    if target.builder_name == 'PythonTest':
        logger.debug('PythonTest target: {}. not using cache.', target.name)
        return True, False
    # read the testing cache.
    try:
        with open(join(cache_dir, 'tested.json'), 'r') as tested_file:
            target.tested = json.loads(tested_file.read())
            test_key = target.test_hash(build_context)
            return True, (target.tested.get(test_key) is not None)
    except json.decoder.JSONDecodeError:
        logger.warning("Got JsonDecodeError when trying to read cache of "
                       "test {}. There is probably a corrupted file. "
                       "Deleting the cache and not using it.", target.name)
        try:
            os.remove(join(cache_dir, 'tested.json'))
        except FileNotFoundError:
            pass
        return True, False


def copy_artifact(src_path: str, artifact_hash: str, conf: Config):
    """Copy the artifact at `src_path` with hash `artifact_hash` to artifacts
       cache dir.

    If an artifact already exists at that location, it is assumed to be
    identical (since it's based on hash), and the copy is skipped.

    TODO: pruning policy to limit cache size.
    """
    cache_dir = conf.get_artifacts_cache_dir()
    if not isdir(cache_dir):
        makedirs(cache_dir, exist_ok=True)
    cached_artifact_path = join(cache_dir, artifact_hash)
    if isfile(cached_artifact_path) or isdir(cached_artifact_path):
        logger.debug('Skipping copy of existing cached artifact {} -> {}',
                     src_path, cached_artifact_path)
        return
    abs_src_path = join(conf.project_root, src_path)
    logger.debug('Caching artifact {} under {}',
                 abs_src_path, cached_artifact_path)
    shutil.copy(abs_src_path, cached_artifact_path)


def restore_artifact(src_path: str, artifact_hash: str, conf: Config):
    """Restore the artifact whose hash is `artifact_hash` to `src_path`.

    Return True if cached artifact is found, valid, and restored successfully.
    Otherwise return False.
    """
    cache_dir = conf.get_artifacts_cache_dir()
    if not isdir(cache_dir):
        return False
    cached_artifact_path = join(cache_dir, artifact_hash)
    if isfile(cached_artifact_path) or isdir(cached_artifact_path):
        # verify cached item hash matches expected hash
        actual_hash = hash_tree(cached_artifact_path)
        if actual_hash != artifact_hash:
            logger.warning(
                'Cached artifact {} expected hash {} != actual hash {}',
                src_path, artifact_hash, actual_hash)
            rmnode(cached_artifact_path)
            return False
        # if something exists in src_path, check if it matches the cached item
        abs_src_path = join(conf.project_root, src_path)
        if isfile(abs_src_path) or isdir(abs_src_path):
            existing_hash = hash_tree(src_path)
            if existing_hash == artifact_hash:
                logger.debug('Existing artifact {} matches cached hash {}',
                             src_path, artifact_hash)
                return True
            logger.debug('Replacing existing artifact {} with cached one',
                         src_path)
            rmnode(abs_src_path)
        logger.debug('Restoring cached artifact {} to {}',
                     artifact_hash, src_path)
        makedirs(dirname(abs_src_path), exist_ok=True)
        shutil.copy(cached_artifact_path, abs_src_path)
        return True
    logger.debug('No cached artifact for {} with hash {}',
                 src_path, artifact_hash)
    return False


def save_target_in_global_cache(target: Target, build_context, cache_dir,
                                artifacts_desc):
    target_hash = target.hash(build_context)
    build_context.global_cache.create_target_cache(target_hash)
    build_context.global_cache.upload_summary(target_hash,
                                              join(cache_dir, 'summary.json'))
    build_context.global_cache.upload_artifacts_meta(
        target_hash, join(cache_dir, 'artifacts.json'))
    build_context.global_cache.upload_artifacts(
        get_artifacts_hashes(artifacts_desc),
        build_context.conf.get_artifacts_cache_dir())


def save_target_in_cache(target: Target, build_context):
    """Save `target` to build cache for future reuse.

    The target hash is used to determine its cache location,
    where the target metadata and artifacts metadata are seriazlied to JSON.
    In addition, relevant artifacts produced by the target are copied under
    the artifacts cache dir by their content hash.

    TODO: pruning policy to limit cache size.
    """
    cache_dir = build_context.conf.get_cache_dir(target, build_context)
    if isdir(cache_dir):
        rmtree(cache_dir)
    makedirs(cache_dir, exist_ok=True)
    logger.debug('Saving target metadata in cache under {}', cache_dir)
    # write target metadata
    with open(join(cache_dir, 'target.json'), 'w') as meta_file:
        meta_file.write(target.json(build_context))
    # copy artifacts to artifact cache by hash
    artifacts = target.artifacts.get_all()
    artifact_hashes = {}
    for artifact_type, artifact_map in artifacts.items():
        if artifact_type in (AT.docker_image,):
            continue
        for dst_path, src_path in artifact_map.items():
            artifact_hashes[dst_path] = hash_tree(src_path)
            # not caching "app" artifacts, since they're part
            # of the source tree
            if artifact_type not in _NO_CACHE_TYPES:
                copy_artifact(src_path, artifact_hashes[dst_path],
                              build_context.conf)
    # serialize target artifacts metadata + hashes
    artifacts_desc = {
        artifact_type.name:
        [{'dst': dst_path, 'src': src_path,
          'hash': artifact_hashes.get(dst_path),
          # We also save the permissions because when downloading from gs we
          # don't get the files with the right permissions.
          'permissions': os.stat(src_path).st_mode
            if artifact_type not in _NO_CACHE_TYPES else None}
         for dst_path, src_path in artifact_map.items()]
        for artifact_type, artifact_map in artifacts.items()
    }
    with open(join(cache_dir, 'artifacts.json'), 'w') as artifacts_meta_file:
        artifacts_meta_file.write(json.dumps(artifacts_desc, indent=4,
                                             sort_keys=True))
    # copying the summary dict so I can modify it without mutating the target
    summary = dict(target.summary)
    summary['name'] = target.name
    summary['artifacts_hash'] = hash_tree(join(cache_dir, 'artifacts.json'))
    if summary.get('created') is None:
        summary['created'] = time()
    write_summary(summary, cache_dir)

    if build_context.global_cache \
            and build_context.conf.upload_to_global_cache:
        try_use_global_cache(build_context, partial(
            save_target_in_global_cache, target, build_context, cache_dir,
            artifacts_desc), 'an error occurred while trying to upload '
                             'target {} to global cache'.format(target.name))


def save_test_in_cache(target: Target, build_context) -> bool:
    """Save `target` testing to build cache for future reuse.

    The target hash is used to determine its cache location,
    where the target testing information is seriazlied to JSON.
    """
    if not target.tested:
        return True
    cache_dir = build_context.conf.get_cache_dir(target, build_context)
    if not isdir(cache_dir):
        logger.debug('Cannot cache test {} - build cache is missing',
                     target.name)
        return False
    tested_path = join(cache_dir, 'tested.json')
    with open(tested_path, 'w') as tested_file:
        tested_file.write(json.dumps(target.tested, indent=4, sort_keys=True))
    target_hash = target.hash(build_context)
    if build_context.global_cache \
            and build_context.conf.upload_tests_to_global_cache:
        try_use_global_cache(
            build_context,
            partial(build_context.global_cache.upload_test_cache, target_hash,
                    tested_path),
            'an error occurred while trying to upload test of target {} to '
            'global cache'.format(target.name))
    return True


def try_use_global_cache(build_context, func, error_msg):
    """
    Checks if should use global cache (by num of failures),
    if there was a failure, increment number of global cache failures.

    :param build_context: contains the global cache
    :param func: the function that uses the global cache
    :param error_msg: error msg to show for a failure.
    :return: return value of func if it worked, False otherwise.
    """
    if build_context.global_cache_failures < MAX_FAILS_FROM_GLOBAL:
        try:
            return func()
        except Exception as e:
            logger.warning(error_msg)
            logger.warning(str(e))
            build_context.global_cache_failures += 1
            return False



# -*- coding: utf-8 -*-

# Copyright 2016 Resonai Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
yabt Extend
~~~~~~~~~~~~~~~~~~~~~~~

:author: Itamar Ostricher
"""


from collections import defaultdict, namedtuple, OrderedDict
from enum import Enum
from functools import partial, wraps
import pkg_resources

from ostrich.utils.collections import listify

from .logging import make_logger


logger = make_logger(__name__)


class Empty(type):
    pass


PropType = Enum('PropType', """str
                               numeric
                               bool
                               list
                               dict
                               StrList
                               TargetName
                               Target
                               TargetList
                               File
                               FileList

                               untyped
                               """)


ArgSpec = namedtuple('ArgSpec', 'type default')


def evaluate_arg_spec(arg_spec):
    arg_type = PropType.untyped
    def_val = Empty
    if isinstance(arg_spec, tuple):
        if len(arg_spec) == 2:
            arg_name, second = arg_spec
            if isinstance(second, PropType):
                arg_type = second
            else:
                def_val = second
        elif len(arg_spec) == 3:
            arg_name, arg_type, def_val = arg_spec
    else:
        arg_name = arg_spec
    # TODO(itamar): better errors than asserts
    assert isinstance(arg_name, str)
    assert isinstance(arg_type, PropType)
    # also check validity of def_val?
    return arg_name, ArgSpec(arg_type, def_val)


INJECTED_ARGS = frozenset((
    'build_params', 'deps', 'cachable', 'license', 'attempts'
    'packaging_params', 'policies', 'runtime_params',
))


class Builder:

    def __init__(self):
        self.sig = None
        self.func = None
        self.test_func = None
        self.cache_json_func = None
        self.docstring = None
        self.min_positional_args = 1  # the `name`

    def register_sig(self, builder_name: str, sig: list, docstring: str,
                     cachable: bool=True, attempts=1):
        """Register a builder signature & docstring for `builder_name`.

        The input for the builder signature is a list of "sig-spec"s
        representing the builder function arguments.

        Each sig-spec in the list can be:
        1. A string. This represents a simple untyped positional argument name,
            with no default value.
        2. A 1-tuple with one string element. Same as #1.
        3. A 2-tuple with ('arg-name', arg_type). This represents a typed
            positional argument, if arg_type is an instance of PropType enum.
        4. A 2-tuple with ('arg-name', default_value). This represents an
            un-typed keyword argument with a default value.
        5. A 3-tuple with ('arg-name', arg_type, default_value). This
            represents a typed keyword argument with a default value,
            if arg_type is an instance of PropType enum.

        In addition to the args specified in the `sig` list, there are several
        *injected* args:
        1. A positional arg `name` of type TargetName is always the first arg.
        2. A keyword arg `deps` of type TargetList and default value `None`
            (or empty list) is always the first after all builder args.
        3. A keyword arg `cachable` of type bool and default value taken from
           the signature registration call (`cachable` arg).
        4. A keyword arg `license` of type StrList and default value [].
        5. A keyword arg `policies` of type StrList and default value [].
        6. A keyword arg `packaging_params` of type dict and default value {}
            (empty dict).
        7. A keyword arg `runtime_params` of type dict and default value {}
            (empty dict).
        8. A keyword arg `build_params` of type dict and default value {}
            (empty dict).
        9. A keyword arg `attempts` of type int and default value 1.
        """
        if self.sig is not None:
            raise KeyError('{} already registered a signature!'
                           .format(builder_name))
        self.sig = OrderedDict(name=ArgSpec(PropType.TargetName, Empty))
        self.docstring = docstring
        kwargs_section = False
        for arg_spec in listify(sig):
            arg_name, sig_spec = evaluate_arg_spec(arg_spec)
            if arg_name in self.sig or arg_name in INJECTED_ARGS:
                raise SyntaxError(
                    "duplicate argument '{}' in function definition"
                    .format(arg_name))
            self.sig[arg_name] = sig_spec
            if sig_spec.default == Empty:
                if kwargs_section:
                    # TODO(itamar): how to give syntax error source annotation?
                    # (see: http://stackoverflow.com/questions/33717804)
                    raise SyntaxError(
                        'non-default argument follows default argument')
                self.min_positional_args += 1
            else:
                kwargs_section = True
        self.sig['deps'] = ArgSpec(PropType.TargetList, None)
        self.sig['cachable'] = ArgSpec(PropType.bool, cachable)
        self.sig['license'] = ArgSpec(PropType.StrList, None)
        self.sig['policies'] = ArgSpec(PropType.StrList, None)
        self.sig['packaging_params'] = ArgSpec(PropType.dict, None)
        self.sig['runtime_params'] = ArgSpec(PropType.dict, None)
        self.sig['build_params'] = ArgSpec(PropType.dict, None)
        self.sig['attempts'] = ArgSpec(PropType.numeric, 1)


class Plugin:

    builders = defaultdict(Builder)
    hooks = {
        'manipulate_target': {},
    }

    @classmethod
    def load_plugins(cls, unused_conf):
        # TODO(itamar): Support config semantics for explicitly enabling /
        # disabling builders, and not just picking up everything that's
        # installed.
        for entry_point in pkg_resources.iter_entry_points('yabt.builders'):
            entry_point.load()
            logger.debug('Loaded builder {0.name} from {0.module_name} '
                         '(dist {0.dist})', entry_point)
        logger.debug('Loaded {} builders', len(cls.builders))
        cls.validate()

    @classmethod
    def get_hooks_for_builder(cls, builder_name: str):
        for hook_name, hook_spec in Plugin.hooks.items():
            if builder_name in hook_spec:
                yield hook_name, hook_spec[builder_name]

    @classmethod
    def validate(cls):
        # TODO(itamar): validate stuff
        # 1. builders are functions with good signatures (name as first arg)
        # 2. hooks belong to existing builders
        pass

    @classmethod
    def remove_builder(cls, builder_name: str):
        """Remove a registered builder `builder_name`.

        No reason to use this except for tests.
        """
        cls.builders.pop(builder_name, None)
        for hook_spec in cls.hooks.values():
            hook_spec.pop(builder_name, None)


def register_builder_sig(
        builder_name, sig=None, docstring=None, cachable: bool=True,
        attempts=1):
    Plugin.builders[builder_name].register_sig(
        builder_name, sig, docstring, cachable, attempts)
    logger.debug('Registered {} builder signature'.format(builder_name))


def register_build_func(builder_name):
    def register_decorator(build_func):
        if Plugin.builders[builder_name].func:
            raise KeyError('{} already registered a build function!'
                           .format(builder_name))
        Plugin.builders[builder_name].func = build_func
        logger.debug('Registered {0} builder signature from '
                     '{1.__module__}.{1.__name__}()', builder_name, build_func)

        @wraps(build_func)
        def builder_wrapper(*args, **kwrags):
            return build_func(*args, **kwrags)
        return builder_wrapper
    return register_decorator


def register_test_func(builder_name):
    def register_decorator(test_func):
        if Plugin.builders[builder_name].test_func:
            raise KeyError('{} already registered a test function!'
                           .format(builder_name))
        Plugin.builders[builder_name].test_func = test_func
        logger.debug('Registered {0} builder signature from '
                     '{1.__module__}.{1.__name__}()', builder_name, test_func)

        @wraps(test_func)
        def tester_wrapper(*args, **kwrags):
            return test_func(*args, **kwrags)
        return tester_wrapper
    return register_decorator


def register_cache_json_func(builder_name):
    def register_decorator(cache_json_func):
        if Plugin.builders[builder_name].cache_json_func:
            raise KeyError('{} already registered a cache_json function!'
                           .format(builder_name))
        Plugin.builders[builder_name].cache_json_func = cache_json_func
        logger.debug('Registered {0} builder signature from '
                     '{1.__module__}.{1.__name__}()', builder_name,
                     cache_json_func)

        @wraps(cache_json_func)
        def cache_json_wrapper(*args, **kwrags):
            return cache_json_func(*args, **kwrags)
        return cache_json_wrapper
    return register_decorator


def _register_hook(hook_name, builder_name):
    def register_decorator(hook_func):
        assert hook_name in Plugin.hooks
        Plugin.hooks[hook_name][builder_name] = hook_func
        logger.debug('Registered {0} hook for {1} builder from '
                     '{2.__module__}.{2.__name__}()',
                     hook_name, builder_name, hook_func)

        @wraps(hook_func)
        def hook_wrapper(*args, **kwargs):
            return hook_func(*args, **kwargs)
        return hook_wrapper
    return register_decorator


register_manipulate_target_hook = partial(_register_hook, 'manipulate_target')



# -*- coding: utf-8 -*-

"""
yabt setup

:copyright: (c) 2016 Resonai by Itamar Ostricher
:license: MIT, see LICENSE for more details.
"""

from os import path

from setuptools import setup, find_packages
import yabt


def get_readme():
    """Read and return the content of the project README file."""
    base_dir = path.abspath(path.dirname(__file__))
    with open(path.join(base_dir, 'README.md'), encoding='utf-8') as readme_f:
        return readme_f.read()


setup(
    name='ybt',
    version=yabt.__version__,
    author=yabt.__author__,
    author_email='yabt@resonai.com',
    url='https://github.com/resonai/ybt',
    description=yabt.__oneliner__,
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'ybt = yabt.yabt:main'
        ],
        'yabt.builders': [
            'Alias = yabt.builders.alias',
            'AptGroup = yabt.builders.apt',
            'AptPackage = yabt.builders.apt',
            'AptRepository = yabt.builders.apt',
            'CppApp = yabt.builders.cpp',
            'CppLib = yabt.builders.cpp',
            'CppProg = yabt.builders.cpp',
            'CppGTest = yabt.builders.cpp',
            'CustomInstaller = yabt.builders.custom_installer',
            'DockerImage = yabt.builders.docker',
            'ExtCommand = yabt.builders.extcommand',
            'ExtDockerImage = yabt.builders.docker',
            'FileGroup = yabt.builders.filegroup',
            'GemPackage = yabt.builders.ruby',
            'GoApp = yabt.builders.golang',
            'GoProg = yabt.builders.golang',
            'Grunt = yabt.builders.grunt',
            'NpmPackage = yabt.builders.nodejs',
            'PythonApp = yabt.builders.python',
            'PythonPackage = yabt.builders.python',
            'Python = yabt.builders.python',
            'PythonTest = yabt.builders.python',
            'TargetGroup = yabt.builders.targetgroup',
            'Proto = yabt.builders.proto',

            'DepTester = yabt.builders.fortests',
        ],
        'yabt.scm': [
            'git = yabt.scm_providers.git',
        ]
    },
    install_requires=[
        'argcomplete',
        'colorama',
        'ConfigArgParse',
        'GitPython',
        'munch',
        'networkx>=2.0',
        'ostrichlib',
        'requests>=2.18.0',
        'scandir',
        'google-cloud-storage',
    ],
    setup_requires=['pytest-runner'],
    extras_require={
        'test': ['pytest', 'pytest-cov', 'pytest-flake8'],
    },
    zip_safe=True,
    license='Apache License, Version 2.0',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development :: Build Tools',
    ]
)
